/***********************************************************************
*                                                                      *
*               This software is part of the ast package               *
*          Copyright (c) 1985-2011 AT&T Intellectual Property          *
*          Copyright (c) 2020-2023 Contributors to ksh 93u+m           *
*                      and is licensed under the                       *
*                 Eclipse Public License, Version 2.0                  *
*                                                                      *
*                A copy of the License is available at                 *
*      https://www.eclipse.org/org/documents/epl-2.0/EPL-2.0.html      *
*         (with md5 checksum 84283fa8859daf213bdda5a9f8d1be1d)         *
*                                                                      *
*                 Glenn Fowler <gsf@research.att.com>                  *
*                  David Korn <dgk@research.att.com>                   *
*                   Phong Vo <kpv@research.att.com>                    *
*                  Martijn Dekker <martijn@inlv.org>                   *
*                                                                      *
***********************************************************************/
/*
 * Glenn Fowler
 * AT&T Research
 *
 * close a proc opened by procopen()
 * otherwise exit() status of process is returned
 */

#include "proclib.h"

int
procclose(Proc_t* p)
{
	int	pid;
	int	flags = 0;
	int	status = -1;

	if (p)
	{
		if (p->rfd >= 0)
			close(p->rfd);
		if (p->wfd >= 0 && p->wfd != p->rfd)
			close(p->wfd);
		if (p->flags & PROC_ORPHAN)
			status = 0;
		else
		{
			if (p->flags & PROC_ZOMBIE)
			{
				/*
				 * process may leave a zombie behind
				 * give it a chance to do that but
				 * don't hang waiting for it
				 */

				flags |= WNOHANG;
				sleep(1);
			}
			if (!(p->flags & PROC_FOREGROUND))
				sigcritical(SIG_REG_EXEC|SIG_REG_PROC);
			while ((pid = waitpid(p->pid, &status, flags)) == -1 && errno == EINTR);
			if (pid != p->pid && (flags & WNOHANG))
				status = 0;
			if (!(p->flags & PROC_FOREGROUND))
				sigcritical(0);
			else
			{
				if (p->sigint != SIG_IGN)
					signal(SIGINT, p->sigint);
				if (p->sigquit != SIG_IGN)
					signal(SIGQUIT, p->sigquit);
#if _lib_sigprocmask
				sigprocmask(SIG_SETMASK, &p->mask, NULL);
#elif _lib_sigsetmask
				sigsetmask(p->mask);
#else
				if (p->sigchld != SIG_DFL)
					signal(SIGCHLD, p->sigchld);
#endif
			}
			status = status == -1 ?
				 EXIT_QUIT :
				 WIFSIGNALED(status) ?
				 EXIT_TERM(WTERMSIG(status)) :
				 EXIT_CODE(WEXITSTATUS(status));
		}
		procfree(p);
	}
	else
		status = errno == ENOENT ? EXIT_NOTFOUND : EXIT_NOEXEC;
	return status;
}
